# Battle for East Asia Scenario
# Civilization 4 (c) 2005 Firaxis Games

# Created by -	Jon 'Trip' Shafer
# Have fun!

from CvPythonExtensions import *
import sys
import Popup as PyPopup
from PyHelpers import PyPlayer
import pickle
import CvEventManager
from CvScreenEnums import *
from PyHelpers import *
import CvUtil

# globals
gc = CyGlobalContext()
localText = CyTranslator()

DefaultUnitAI = UnitAITypes.NO_UNITAI

class CvBFEAEvents(CvEventManager.CvEventManager):
	
	def __init__(self):
		
		CvEventManager.CvEventManager.__init__(self)
		
		# Offset value set in WorldBuilder save
		self.iTurnOffset = 4
		
		self.bNukesEnabled = false
		# First turn Nukes can be unlocked
		self.iNukesEnabledTurn = 30#30
		# Score one of the two main teams has to go under to enable nukes
		self.iNukeScoreThreshold = 280#275
		
		# First turn the event rolls to see if these guys go to war
		self.iChinaJapanWarTurn = 30#30
		# Modifies turn # to get odds of going to war - lower number == higher chance
		self.iChinaJapanWarDivisor = 4#4
		# Debug means of increasing odds of Japan-China war
		self.iChinaJapanWarMod = 0#0
		
		# Only display the rebels event message once
		self.bShownRussianRebelText = false
		# The higher this number the more often the Russians have to deal with rebels (compared to Russia's score)
		self.iRussianRebelsSpawnFactor = 170#170
		self.iCitySpawnUnitCounterDefault = 0
		# Number of turns between possible rebel spawning
		self.iCitySpawnUnitCounterMax = 6#6
		self.iUnitSpawnRange = 3#3
		# Debug means of increasing odds of Rebel spawning
		self.iRussianRebelSpawnMod = 0#0
		
		self.bShownRussianReinforceText = false
		self.iRussianReinforcementThreshold = 35#35
		self.iRussianGoldAmount = 50#50
		self.iRussianReinforceCounter = 10#10
		self.aiRussianReinforceLandPlot = [0, 79]
		
		self.bShownAmericanReinforceText = false
		self.iAmericanReinforcementThreshold = 40#40
		self.iAmericanGoldAmount = 200#200
		self.iAmericanReinforceCounter = 8#8
		self.aiAmericanReinforceLandPlot = [56, 43]
		self.aiAmericanReinforceWaterPlot = [62, 34]
		
		self.bShownUnitedNationReinforceText = false
		self.iUnitedNationsReinforcementThreshold = 40
		# Determines how many units the UN gets as reinforcements in India - lower number = more units (compared with China's score)
		self.iUNReinforceChinaFactor = 330#330
		self.iUnitedNationsGoldAmount = 50#50
		self.iUnitedNationsReinforceCounter = 8#8
		self.aiUnitedNationsReinforceLandPlot = [0, 26]
		
		self.iChinaTeamID =	0
		self.iAlliesTeamID =	1
		
		self.iChinaID =		0
		self.iUSAID =		1
		self.iRussiaID =	2
		self.iJapanID =		3
		self.iUnitedNationsID =	4
		self.iVietnamID =	5
		self.iLaosID =		6
		self.iCambodiaID =	7
		self.iBurmaID =		8
		self.iThailandID =	9
		self.iNorthKoreaID =	10
		self.iSouthKoreaID =	11
		self.iMongoliaID =	12
		self.iNumPlayers =      13
		self.iBarbarianID =	18
		
		self.szEventText = ""
		self.szResultText = ""
		
	def turnChecker(self, iTurn):
		
		# XXX - Temp for use while debugging/reloading
		self.initValues()
		
		pChinese = gc.getPlayer(self.iChinaID)
		pUSA = gc.getPlayer(self.iUSAID)
		
		self.szGameDate = CyGameTextMgr().getTimeStr(iTurn, false)
		
		# Check for Japanese/Chinese Declaration of War
		self.checkForNewWars(iTurn)
		
		# City spawning rebel stuff
		self.checkForUnitSpawning()
		self.alterCitySpawnUnitCounters()
		
		# Check for enabling of Nukes
		self.checkForNukes(iTurn)
		
		# Russian reinforcements every [self.iRussianhReinforceCounter] turns
		if ((iTurn - self.iTurnOffset) > 0 and (iTurn - self.iTurnOffset) % self.iRussianReinforceCounter == 0):
			self.addRussianReinforcements(iTurn)
		
		# American reinforcements every [self.iAmericanReinforceCounter] turns
		if ((iTurn - self.iTurnOffset) > 0
		    and (iTurn - self.iTurnOffset) % self.iAmericanReinforceCounter == 0):
			
			self.addAmericanReinforcements(iTurn)
		
		# UnitedNations reinforcements every [self.iUnitedNationsReinforceCounter] turns
		if ((iTurn - self.iTurnOffset) > 0
		    and (iTurn - self.iTurnOffset) % self.iUnitedNationsReinforceCounter == 0):
			
			self.addUnitedNationsReinforcements(iTurn)
		
	# Determine the start-game state
	def setupGame(self):
		
		# Temporary Measure to turn off 'Feats'
		CyMessageControl().sendPlayerOption(PlayerOptionTypes.PLAYEROPTION_ADVISOR_POPUPS, false)
		
		self.initValues()
		
		pChina = gc.getPlayer(self.iChinaID)
		pChinaTeam = gc.getTeam(pChina.getTeam())
		
		pUSA = gc.getPlayer(self.iUSAID)
		pUSATeam = gc.getTeam(pUSA.getTeam())
		
		for iPlayerLoop in range(self.iNumPlayers):
			
			pPlayer = gc.getPlayer(iPlayerLoop)
			pyPlayer = PyPlayer(iPlayerLoop)
			
			# Set player default research to 80%
			pPlayer.setCommercePercent(CommerceTypes.COMMERCE_RESEARCH, 90)
			
			# Loop through player's cities
			apCityList = pyPlayer.getCityList()
			
			for pCityLoop in apCityList:
				
				# Initialize City
				self.initCity(pCityLoop)
				
	def initValues(self):
		
		self.iDestroyerID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_DESTROYER')
		self.iBattleshipID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_BATTLESHIP')
		self.iCarrierID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_CARRIER')
		self.iTransportID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_TRANSPORT')
		
		self.iModernArmorID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_MODERN_ARMOR')#83
		self.iAdvancedArmorID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_ADVANCED_ARMOR')
		self.iInfantryIFVID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_INFANTRY_IFV')
		self.iAdvancedIFVID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_ADVANCED_IFV')
		self.iInfantryID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_INFANTRY')
		self.iMarineID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_MARINE')
		self.iArtilleryID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_ARTILLERY')
		self.iMobileArtilleryID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_MOBILE_ARTILLERY')
		self.iSAMInfantryID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_SAM_INFANTRY')
		self.iMobileSAMID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_MOBILE_SAM')
		self.iGunshipID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_GUNSHIP')
		self.iAdvancedGunshipID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_ADVANCED_GUNSHIP')
		self.iJetFighterID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_JET_FIGHTER')
		self.iAdvancedFighterID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_ADVANCED_FIGHTER')
		self.iStrikeAircraftID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_STRIKE_AIRCRAFT')
		
		self.iType98TankID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_TYPE_98_TANK')
		self.iType12TankID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_TYPE_12_TANK')
		self.iType86IFVID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_TYPE_86_IFV')
		self.iType06IFVID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_TYPE_06_IFV')
		self.iHQ15MobileSAMID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_HQ-15_MOBILE_SAM')
		self.iQ6FighterID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_Q-6_FIGHTER')
		self.iJ12FighterID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_J-12_FIGHTER')
		
		self.iM1A2AbramsTankID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_M1A2_ABRAMS_TANK')
		self.iM1A5AbramsTankID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_M1A5_ABRAMS_TANK')
		self.iM2BradleyIFVID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_M2_BRADLEY_IFV')
		self.iM2A4BradleyIFVID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_M2A4_BRADLEY_IFV')
		self.iAmericanInfantryID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_AMERICAN_INFANTRY')
		self.iPaladinMobileArtilleryID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_PALADIN_MOBILE_ARTILLERY')
		self.iPatriotMobileSAMID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_PATRIOT_MOBILE_SAM')
		self.iF16FighterID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_F-16_FIGHTER')
		self.iF35FighterID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_F-35_FIGHTER')
		self.iB2StealthBomberID =	CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'BFEA_UNIT_B-2_STEALTH_BOMBER')
		
		self.iNuclearProliferationID =	    CvUtil.findInfoTypeNum(gc.getTechInfo,gc.getNumTechInfos(),'BFEA_TECH_NUCLEAR_PROLIFERATION')
		
		for iPlayerLoop in range(self.iNumPlayers):
			
			pPlayer = gc.getPlayer(iPlayerLoop)
			pyPlayer = PyPlayer(iPlayerLoop)
			# Loop through player's cities
			apCityList = pyPlayer.getCityList()
			
			for pCityLoop in apCityList:
				
				# Initialize City
				self.initCity(pCityLoop)
				
	def initCity(self, pCity):
		
		aiCityXY = [pCity.getX(), pCity.getY()]
		
		# This will become the city "ID" within the scenario used to identify it later (in case it's renamed or something)
		iCityCount = -1
		
		if (aiCityXY == [22, 76]):
			self.iYAKUTSK = iCityCount = 0
		elif (aiCityXY == [1, 73]):
			self.iKRASNOYARSK = iCityCount = 1
		elif (aiCityXY == [10, 67]):
			self.iIRKUTSK = iCityCount = 2
		elif (aiCityXY == [41, 55]):
			self.iVLADIVOSTOK = iCityCount = 3
		
		elif (aiCityXY == [11, 62]):
			self.iULAANBATAAR = iCityCount = 4
		
		elif (aiCityXY == [52, 54]):
			self.iSAPPORO = iCityCount = 5
		elif (aiCityXY == [56, 43]):
			self.iTOKYO = iCityCount = 6
		elif (aiCityXY == [51, 42]):
			self.iOSAKA = iCityCount = 7
		elif (aiCityXY == [47, 38]):
			self.iKAGOSHIMA = iCityCount = 8
		
		elif (aiCityXY == [40, 45]):
			self.iSEOUL = iCityCount = 9
		
		elif (aiCityXY == [34, 58]):
			self.iHARBIN = iCityCount = 10
		elif (aiCityXY == [25,50 ]):
			self.iBEIJING = iCityCount = 11
		elif (aiCityXY == [35, 38]):
			self.iSHANGHAI = iCityCount = 12
		elif (aiCityXY == [18, 26]):
			self.iKUNMING = iCityCount = 13
		elif (aiCityXY == [39, 26]):
			self.iTAIPEI = iCityCount = 14
		
		elif (aiCityXY == [0, 25]):
			self.iDHAKA = iCityCount = 15
		elif (aiCityXY == [40, 12]):
			self.iMANILA = iCityCount = 16
		
		elif (aiCityXY == [6, 15]):
			self.iRANGOON = iCityCount = 17
		
		elif (aiCityXY == [15, 17]):
			self.iVIANGCHAN = iCityCount = 17
		
		elif (aiCityXY == [19, 22]):
			self.iHANOI = iCityCount = 18
		elif (aiCityXY == [23, 6]):
			self.iHOCHIMINHCITY = iCityCount = 20
		
		elif (aiCityXY == [20, 7]):
			self.iPHNOMPENH = iCityCount = 21
		
		elif (aiCityXY == [14, 10]):
			self.iBANGKOK = iCityCount = 22
		
		elif (aiCityXY == [37, 48]):
			self.iPYONGYANG = iCityCount = 23
		
		# NON-VP CITIES
		elif (aiCityXY == [7, 72]):
			self.iBRATSK = iCityCount = 24
		elif (aiCityXY == [5, 26]):
			self.iIMPHAL = iCityCount = 25
		
		# Set default script data manually since we need defaults for all values in the array
		aScriptData = [iCityCount, self.iCitySpawnUnitCounterDefault]
		pCity.setScriptData(pickle.dumps(aScriptData))
		
###########################################################################################
####################################### MISC EVENTS #######################################
###########################################################################################
			
	def checkForNewWars(self, iTurn):
		
		# If we're at least iChinaJapanWarTurn (30) turns in then there's an increasing chance of China and Japan going to war
		if (iTurn > self.iChinaJapanWarTurn):
			pChina = gc.getPlayer(self.iChinaID)
			iChinaTeam = pChina.getTeam()
			pChinaTeam = gc.getTeam(iChinaTeam)
			pJapan = gc.getPlayer(self.iJapanID)
			iJapanTeam = pJapan.getTeam()
			pJapanTeam = gc.getTeam(iJapanTeam)
			
			if (pChinaTeam.isAtWar(iJapanTeam)):
				return
			
			iRand = self.getRand(99)

			# Apply factor to turn to get random DOW
			if ((iTurn / self.iChinaJapanWarDivisor) < iRand + self.iChinaJapanWarMod):
				if (pChina.isHuman()):
					pJapanTeam.declareWar(iChinaTeam, false)
					
					self.szEventText = localText.getText("TXT_KEY_BFEA_JAPAN_DOW_TEXT", ())
					self.szResultText = localText.getText("TXT_KEY_BFEA_JAPAN_DOW_RESULT", ())
					
					self.displayEventText()
					
				elif (not pChina.isHuman()):
					pChinaTeam.declareWar(iJapanTeam, false)
					
					self.szEventText = localText.getText("TXT_KEY_BFEA_CHINA_DOW_TEXT", ())
					self.szResultText = localText.getText("TXT_KEY_BFEA_CHINA_DOW_RESULT", ())
					
					self.displayEventText()
			    
		return
		
	def alterCitySpawnUnitCounters(self):
		
		# Loop through all players' cities
		for iPlayerLoop in range(self.iNumPlayers):
			
			pPlayer = PyPlayer(iPlayerLoop)
			apCityList = pPlayer.getCityList()
			
			for pCity in apCityList:
				
				iCityID = pCity.getID()
				# Reduce CitySpawnUnit counter by 1 in applicable cities
				if (self.getCitySpawnUnitCounter(iPlayerLoop, iCityID) > 0):
					
					self.changeCitySpawnUnitCounter(iPlayerLoop, iCityID, -1)
					
	def checkForUnitSpawning(self):
		
		iRebelSpawnChance = int(self.iRussianRebelsSpawnFactor / (gc.getGame().getPlayerScore(self.iRussiaID) * 1.0) * 15) + self.iRussianRebelSpawnMod
#		print("iRebelSpawnChance: %d" %(iRebelSpawnChance))
		# Loop through all Russian cities
		pyPlayer = PyPlayer(self.iRussiaID)
		apCityList = pyPlayer.getCityList()
		
		for pCityLoop in apCityList:
			iCityID = pCityLoop.getID()
			iCityX = pCityLoop.getX()
			iCityY = pCityLoop.getY()
			
			# If this city has spawned a unit recently then no more units
			if (self.getCitySpawnUnitCounter(self.iRussiaID, iCityID) == 0):
				
				iUnitSpawnRoll = self.getRand(99)
#				print("iUnitSpawnRoll: %d" %(iUnitSpawnRoll))
				# Create Infantry for the Barbarians
				if (iUnitSpawnRoll < iRebelSpawnChance):
					
					aiSpawnPlot = self.findUnitPlacementPlot(self.iBarbarianID, iCityX, iCityY, true, true, self.iUnitSpawnRange)
#					print("Spawn Plot: %d, %d" %(aiSpawnPlot[0], aiSpawnPlot[1]))
					if (aiSpawnPlot == "Oh Snap"):
						return
					else:
						pBarbarians = gc.getPlayer(self.iBarbarianID)
						
						pBarbarians.initUnit(self.iInfantryID, aiSpawnPlot[0], aiSpawnPlot[1], DefaultUnitAI)
						
						self.setCitySpawnUnitCounter(self.iRussiaID, iCityID, self.iCitySpawnUnitCounterMax)
						
						if (self.bShownRussianRebelText == false):
							
							self.bShownRussianRebelText = true
							
							# Set event text
							self.szEventText = localText.getText("TXT_KEY_BFEA_RUSSIAN_REBELS_TEXT", ())
							self.szResultText = localText.getText("TXT_KEY_BFEA_RUSSIAN_REBELS_RESULT", (pCityLoop.getName(),))
							
							self.displayEventText()
	
	def checkForNukes(self, iTurn):
		
		# If nukes area already enabled then don't do anything
		if (self.bNukesEnabled == true):
			return
		
		# Only enabled after a certain number of turns
		if (iTurn >= self.iNukesEnabledTurn):
			
			iChinaTeam = gc.getPlayer(self.iChinaID).getTeam()
			pChinaTeam = gc.getTeam(iChinaTeam)
			iAlliesTeam = gc.getPlayer(self.iUSAID).getTeam()
			pAlliesTeam = gc.getTeam(iAlliesTeam)
			
			# One team's score must be below a certain level in order to enable nukes
			if (gc.getGame().getTeamScore(iChinaTeam) > self.iNukeScoreThreshold and gc.getGame().getTeamScore(iAlliesTeam) > self.iNukeScoreThreshold):
				return
			
			self.bNukesEnabled = true
			
			# Award Nuclear Proliferation tech to both teams and show event text
			pChinaTeam.setHasTech(self.iNuclearProliferationID, true, self.iChinaID, false, false)
			pAlliesTeam.setHasTech(self.iNuclearProliferationID, true, self.iUSAID, false, false)
			
			if (gc.getGame().getTeamScore(iChinaTeam) < self.iNukeScoreThreshold):
				self.szEventText = localText.getText("TXT_KEY_BFEA_NUKES_ENABLED_CHINA_TEXT", ())
				self.szResultText = localText.getText("TXT_KEY_BFEA_NUKES_ENABLED_RESULT", ())
				
			if (gc.getGame().getTeamScore(iAlliesTeam) < self.iNukeScoreThreshold):
				self.szEventText = localText.getText("TXT_KEY_BFEA_NUKES_ENABLED_ALLIES_TEXT", ())
				self.szResultText = localText.getText("TXT_KEY_BFEA_NUKES_ENABLED_RESULT", ())
			
			self.displayEventText()
		
	def addRussianReinforcements(self, iTurn):
		
		pRussianPlayer = gc.getPlayer(self.iRussiaID)
		
		# Increase coffers (can't be going bankrupt now)
		pRussianPlayer.changeGold(self.iRussianGoldAmount)
		
		# Determine force list based upon what turn it is
		
		if (iTurn < self.iRussianReinforcementThreshold + self.iTurnOffset):
			iNumInfantry =		3
			iInfantryChance =	50
			iNumPossInfantry =	2
			
			iNumInfantryIFV =	0
			iInfantryIFVChance =	25
			iNumPossInfantryIFV =	2
			
			iNumAdvancedIFV =	0
			iInfantryIFVChance =	0
			iNumPossAdvancedIFV =	0
			
			iNumModernArmor =	0
			iModernArmorChance =	50
			iNumPossModernArmor =	1
			
			iNumAdvancedArmor =	0
			iAdvancedArmorChance =	0
			iNumPossAdvancedArmor =	0
			
			iNumArtillery =		1
			iArtilleryChance =	50
			iNumPossArtillery =	1
			
			iNumMobileArtillery =		0
			iMobileArtilleryChance =	40
			iNumPossMobileArtillery =	1
			
			iNumSAMInfantry =	1
			iSAMInfantryChance =	50
			iNumPossSAMInfantry =	1
			
			iNumMobileSAM =		0
			iMobileSAMChance =	50
			iNumPossMobileSAM =	1
			
			iNumGunship =		0
			iGunshipChance =	50
			iNumPossGunship =	2
			
			iNumAdvancedGunship =		0
			iAdvancedGunshipChance =	0
			iNumPossAdvancedGunship =	0
			
			iNumJetFighter =	0
			iJetFighterChance =	50
			iNumPossJetFighter =	2
			
			iNumAdvancedFighter =		0
			iAdvancedFighterChance =	0
			iNumPossAdvancedFighter =	0
			
			iNumStrikeAircraft =	0
			iStrikeAircraftChance =	50
			iNumPossStrikeAircraft =2
			
		else:
			iNumInfantry =		2
			iInfantryChance =	50
			iNumPossInfantry =	1
			
			iNumInfantryIFV =	1
			iInfantryIFVChance =	35
			iNumPossInfantryIFV =	2
			
			iNumAdvancedIFV =	0
			iInfantryIFVChance =	50
			iNumPossAdvancedIFV =	2
			
			iNumModernArmor =	1
			iModernArmorChance =	50
			iNumPossModernArmor =	2
			
			iNumAdvancedArmor =	0
			iAdvancedArmorChance =	50
			iNumPossAdvancedArmor =	1
			
			iNumArtillery =		1
			iArtilleryChance =	50
			iNumPossArtillery =	1
			
			iNumMobileArtillery =		1
			iMobileArtilleryChance =	40
			iNumPossMobileArtillery =	1
			
			iNumSAMInfantry =		0
			iSAMInfantryChance =		50
			iNumPossSAMInfantry =		2
			
			iNumGunship =			1
			iGunshipChance =		50
			iNumPossGunship =		1
			
			iNumAdvancedGunship =		0
			iAdvancedGunshipChance =	50
			iNumPossAdvancedGunship =	1
			
			iNumJetFighter =		1
			iJetFighterChance =		30
			iNumPossJetFighter =		1
			
			iNumAdvancedFighter =		0
			iAdvancedFighterChance =	30
			iNumPossAdvancedFighter =	1
			
			iNumStrikeAircraft =		0
			iStrikeAircraftChance =		50
			iNumPossStrikeAircraft =	2
			
		# Determine the number of units to spawn
		for iUnitNumLoop in range(iNumPossInfantry):
			if (self.getRand(99) < iInfantryChance):
				iNumInfantry += 1
		for iUnitNumLoop in range(iNumPossInfantryIFV):
			if (self.getRand(99) < iInfantryIFVChance):
				iNumInfantryIFV += 1
		for iUnitNumLoop in range(iNumPossAdvancedIFV):
			if (self.getRand(99) < iAdvancedIFVChance):
				iNumAdvancedIFV += 1
		for iUnitNumLoop in range(iNumPossModernArmor):
			if (self.getRand(99) < iModernArmorChance):
				iNumModernArmor += 1
		for iUnitNumLoop in range(iNumPossAdvancedArmor):
			if (self.getRand(99) < iAdvancedArmorChance):
				iNumAdvancedArmor += 1
		for iUnitNumLoop in range(iNumPossArtillery):
			if (self.getRand(99) < iArtilleryChance):
				iNumArtillery += 1
		for iUnitNumLoop in range(iNumPossMobileArtillery):
			if (self.getRand(99) < iMobileArtilleryChance):
				iNumMobileArtillery += 1
		for iUnitNumLoop in range(iNumPossSAMInfantry):
			if (self.getRand(99) < iSAMInfantryChance):
				iNumSAMInfantry += 1
		for iUnitNumLoop in range(iNumPossMobileSAM):
			if (self.getRand(99) < iMobileSAMChance):
				iNumMobileSAM += 1
		for iUnitNumLoop in range(iNumPossGunship):
			if (self.getRand(99) < iGunshipChance):
				iNumGunship += 1
		for iUnitNumLoop in range(iNumPossAdvancedGunship):
			if (self.getRand(99) < iAdvancedGunshipChance):
				iNumAdvancedGunship += 1
		for iUnitNumLoop in range(iNumPossJetFighter):
			if (self.getRand(99) < iJetFighterChance):
				iNumJetFighter += 1
		for iUnitNumLoop in range(iNumPossAdvancedFighter):
			if (self.getRand(99) < iAdvancedFighterChance):
				iNumAdvancedFighter += 1
		for iUnitNumLoop in range(iNumPossStrikeAircraft):
			if (self.getRand(99) < iStrikeAircraftChance):
				iNumStrikeAircraft += 1
		
		# Determine a plot for land units to go to, then add them
		
		iLandPlotRange = 3
		
		aiLandUnitsPlot = self.findUnitPlacementPlot(self.iRussiaID, self.aiRussianReinforceLandPlot[0], self.aiRussianReinforceLandPlot[1])
		
		if (aiLandUnitsPlot == "Oh Snap"):
			return
			
		else:
			iX = aiLandUnitsPlot[0]
			iY = aiLandUnitsPlot[1]
			
			for iUnitLoop in range(iNumInfantry):
				pRussianPlayer.initUnit(self.iInfantryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumInfantryIFV):
				pRussianPlayer.initUnit(self.iInfantryIFVID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedIFV):
				pRussianPlayer.initUnit(self.iAdvancedIFVID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumModernArmor):
				pRussianPlayer.initUnit(self.iModernArmorID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedArmor):
				pRussianPlayer.initUnit(self.iAdvancedArmorID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumArtillery):
				pRussianPlayer.initUnit(self.iArtilleryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumMobileArtillery):
				pRussianPlayer.initUnit(self.iMobileArtilleryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumSAMInfantry):
				pRussianPlayer.initUnit(self.iSAMInfantryID, iX, iY, UnitAITypes.UNITAI_COUNTER)
			for iUnitLoop in range(iNumMobileSAM):
				pRussianPlayer.initUnit(self.iMobileSAMID, iX, iY, UnitAITypes.UNITAI_COUNTER)
			for iUnitLoop in range(iNumGunship):
				pRussianPlayer.initUnit(self.iGunshipID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedGunship):
				pRussianPlayer.initUnit(self.iAdvancedGunshipID, iX, iY, UnitAITypes.UNITAI_ATTACK)
		
		# Add aircraft
		aiValidCities = [	self.iKRASNOYARSK,
					self.iBRATSK	]
		
		iCityID = self.determineValidCities(self.iRussiaID, aiValidCities)
		
		if (iCityID > -1):

			iCityX = pRussianPlayer.getCity(iCityID).getX()
			iCityY = pRussianPlayer.getCity(iCityID).getY()
			
			for iUnitLoop in range(iNumJetFighter):
				pRussianPlayer.initUnit(self.iJetFighterID, iCityX, iCityY, UnitAITypes.UNITAI_ATTACK_AIR)
			for iUnitLoop in range(iNumAdvancedFighter):
				pRussianPlayer.initUnit(self.iAdvancedFighterID, iCityX, iCityY, UnitAITypes.UNITAI_ATTACK_AIR)
			for iUnitLoop in range(iNumStrikeAircraft):
				pRussianPlayer.initUnit(self.iStrikeAircraftID, iCityX, iCityY, UnitAITypes.UNITAI_ATTACK_AIR)
		
		# Set event text
		if (self.bShownRussianReinforceText == false):
			self.bShownRussianReinforceText = true
			self.szEventText = localText.getText("TXT_KEY_BFEA_RUSSIA_REINFORCE_TEXT", ())
			self.szResultText = localText.getText("TXT_KEY_BFEA_RUSSIA_REINFORCE_RESULT", ())
			
			self.displayEventText()
			
	def addAmericanReinforcements(self, iTurn):
		
		pAmericanPlayer = gc.getPlayer(self.iUSAID)
		
		# Increase coffers (can't be going bankrupt now)
		pAmericanPlayer.changeGold(self.iAmericanGoldAmount)
		
		# Determine force list based upon what turn it is
		
		if (iTurn < self.iAmericanReinforcementThreshold + self.iTurnOffset):
			iNumAmericanInfantry =		2
			iAmericanInfantryChance =	50
			iNumPossAmericanInfantry =	2
			
			iNumM2BradleyIFV =		1
			iM2BradleyIFVChance =		50
			iNumPossM2BradleyIFV =		1
			
			iNumM2A4BradleyIFV =		0
			iM2A4BradleyIFVChance =		0
			iNumPossM2A4BradleyIFV =	0
			
			iNumM1A2AbramsTank =		0
			iM1A2AbramsTankChance =		50
			iNumPossM1A2AbramsTank =	2
			
			iNumM1A5AbramsTank =		0
			iM1A5AbramsTankChance =		0
			iNumPossM1A5AbramsTank =	0
			
			iNumArtillery =			1
			iArtilleryChance =		50
			iNumPossArtillery =		1
			
			iNumPaladinMobileArtillery =		0
			iPaladinMobileArtilleryChance =		30
			iNumPossPaladinMobileArtillery =	1
			
			iNumSAMInfantry =	1
			iSAMInfantryChance =	50
			iNumPossSAMInfantry =	1
			
			iNumPatriotMobileSAM =		0
			iPatriotMobileSAMChance =	50
			iNumPossPatriotMobileSAM =	0
			
			iNumGunship =		0
			iGunshipChance =	50
			iNumPossGunship =	2
			
			iNumAdvancedGunship =		0
			iAdvancedGunshipChance =	0
			iNumPossAdvancedGunship =	0
			
			iNumB2Bomber =		0
			iB2BomberChance =	30
			iNumPossB2Bomber =	1
			
			iNumCarrier =		1
			iCarrierChance =	50
			iNumPossCarrier =	2
			
			iNumDestroyer =		2
			iDestroyerChance =	50
			iNumPossDestroyer =	2
			
			iNumBattleship =	0
			iBattleshipChance =	50
			iNumPossBattleship =	2
			
			iNumTransport =		2
			iTransportChance =	50
			iNumPossTransport =	2
			
		else:
			iNumAmericanInfantry =		1
			iAmericanInfantryChance =	50
			iNumPossAmericanInfantry =	1
			
			iNumM2BradleyIFV =		2
			iM2BradleyIFVChance =		50
			iNumPossM2BradleyIFV =		1
			
			iNumM2A4BradleyIFV =		1
			iM2A4BradleyIFVChance =		50
			iNumPossM2A4BradleyIFV =	1
			
			iNumM1A2AbramsTank =		1
			iM1A2AbramsTankChance =		50
			iNumPossM1A2AbramsTank =	1
			
			iNumM1A5AbramsTank =		1
			iM1A5AbramsTankChance =		50
			iNumPossM1A5AbramsTank =	2
			
			iNumArtillery =			1
			iArtilleryChance =		50
			iNumPossArtillery =		1
			
			iNumPaladinMobileArtillery =		1
			iPaladinMobileArtilleryChance =		50
			iNumPossPaladinMobileArtillery =	2
			
			iNumSAMInfantry =	0
			iSAMInfantryChance =	50
			iNumPossSAMInfantry =	1
			
			iNumPatriotMobileSAM =		1
			iPatriotMobileSAMChance =	50
			iNumPossPatriotMobileSAM =	2
			
			iNumGunship =		0
			iGunshipChance =	50
			iNumPossGunship =	1
			
			iNumAdvancedGunship =		1
			iAdvancedGunshipChance =	50
			iNumPossAdvancedGunship =	2
			
			iNumB2Bomber =		0
			iB2BomberChance =	50
			iNumPossB2Bomber =	2
			
			iNumCarrier =		1
			iCarrierChance =	50
			iNumPossCarrier =	2
			
			iNumDestroyer =		2
			iDestroyerChance =	50
			iNumPossDestroyer =	2
			
			iNumBattleship =	0
			iBattleshipChance =	50
			iNumPossBattleship =	2
			
			iNumTransport =		2
			iTransportChance =	50
			iNumPossTransport =	2
			
		# Determine the number of units to spawn
		for iUnitNumLoop in range(iNumPossAmericanInfantry):
			if (self.getRand(99) < iAmericanInfantryChance):
				iNumAmericanInfantry += 1
		for iUnitNumLoop in range(iNumPossM2BradleyIFV):
			if (self.getRand(99) < iM2BradleyIFVChance):
				iNumM2BradleyIFV += 1
		for iUnitNumLoop in range(iNumPossM2A4BradleyIFV):
			if (self.getRand(99) < iM2A4BradleyIFVChance):
				iNumM2A4BradleyIFV += 1
		for iUnitNumLoop in range(iNumPossM1A2AbramsTank):
			if (self.getRand(99) < iM1A2AbramsTankChance):
				iNumM1A2AbramsTank += 1
		for iUnitNumLoop in range(iNumPossM1A5AbramsTank):
			if (self.getRand(99) < iM1A5AbramsTankChance):
				iNumM1A5AbramsTank += 1
		for iUnitNumLoop in range(iNumPossArtillery):
			if (self.getRand(99) < iArtilleryChance):
				iNumArtillery += 1
		for iUnitNumLoop in range(iNumPossPaladinMobileArtillery):
			if (self.getRand(99) < iPaladinMobileArtilleryChance):
				iNumPaladinMobileArtillery += 1
		for iUnitNumLoop in range(iNumPossSAMInfantry):
			if (self.getRand(99) < iSAMInfantryChance):
				iNumSAMInfantry += 1
		for iUnitNumLoop in range(iNumPossPatriotMobileSAM):
			if (self.getRand(99) < iPatriotMobileSAMChance):
				iNumPatriotMobileSAM += 1
		for iUnitNumLoop in range(iNumPossGunship):
			if (self.getRand(99) < iGunshipChance):
				iNumGunship += 1
		for iUnitNumLoop in range(iNumPossAdvancedGunship):
			if (self.getRand(99) < iAdvancedGunshipChance):
				iNumAdvancedGunship += 1
				
		for iUnitNumLoop in range(iNumPossCarrier):
			if (self.getRand(99) < iCarrierChance):
				iNumCarrier += 1
		for iUnitNumLoop in range(iNumPossDestroyer):
			if (self.getRand(99) < iDestroyerChance):
				iNumGunship += 1
		for iUnitNumLoop in range(iNumPossBattleship):
			if (self.getRand(99) < iBattleshipChance):
				iNumBattleship += 1
		for iUnitNumLoop in range(iNumPossTransport):
			if (self.getRand(99) < iTransportChance):
				iNumTransport += 1
				
		if (iTurn < 40 + self.iTurnOffset):
			iNumF16Fighter = iNumCarrier * 2
			iNumF35Fighter = 0
			iNumStrikeAircraft = iNumCarrier
		else:
			iNumF16Fighter = iNumCarrier
			iNumF35Fighter = iNumCarrier
			iNumStrikeAircraft = iNumCarrier
		
		# Determine a plot for land units to go to, then add them
		
		iLandPlotRange = 3
		
		aiLandUnitsPlot = self.findUnitPlacementPlot(self.iUSAID, self.aiAmericanReinforceLandPlot[0], self.aiAmericanReinforceLandPlot[1])
		
		if (aiLandUnitsPlot == "Oh Snap"):
			return
			
		else:
			iX = aiLandUnitsPlot[0]
			iY = aiLandUnitsPlot[1]
			
			for iUnitLoop in range(iNumAmericanInfantry):
				pAmericanPlayer.initUnit(self.iAmericanInfantryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumM2BradleyIFV):
				pAmericanPlayer.initUnit(self.iM2BradleyIFVID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumM2A4BradleyIFV):
				pAmericanPlayer.initUnit(self.iM2A4BradleyIFVID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumM1A2AbramsTank):
				pAmericanPlayer.initUnit(self.iM1A2AbramsTankID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumM1A5AbramsTank):
				pAmericanPlayer.initUnit(self.iM1A5AbramsTankID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumArtillery):
				pAmericanPlayer.initUnit(self.iArtilleryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumPaladinMobileArtillery):
				pAmericanPlayer.initUnit(self.iPaladinMobileArtilleryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumPatriotMobileSAM):
				pAmericanPlayer.initUnit(self.iPatriotMobileSAMID, iX, iY, UnitAITypes.UNITAI_COUNTER)
			for iUnitLoop in range(iNumGunship):
				pAmericanPlayer.initUnit(self.iGunshipID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedGunship):
				pAmericanPlayer.initUnit(self.iAdvancedGunshipID, iX, iY, UnitAITypes.UNITAI_ATTACK)
				
		# Find a plot out to sea to put the new American fleet
		
		iWaterPlotRange = 5
		
		aiWaterUnitsPlot = self.findUnitPlacementPlot(self.iUSAID, self.aiAmericanReinforceWaterPlot[0], self.aiAmericanReinforceWaterPlot[1], false)
		
		if (aiWaterUnitsPlot == "Oh Snap"):
			return
			
		else:
			iX = aiWaterUnitsPlot[0]
			iY = aiWaterUnitsPlot[1]
			
			# Add the boats first and the air units will automatically load aboard them - neat!
			for iUnitLoop in range(iNumDestroyer):
				pAmericanPlayer.initUnit(self.iDestroyerID, iX, iY, UnitAITypes.UNITAI_ATTACK_SEA)
			for iUnitLoop in range(iNumBattleship):
				pAmericanPlayer.initUnit(self.iBattleshipID, iX, iY, UnitAITypes.UNITAI_ATTACK_SEA)
			for iUnitLoop in range(iNumCarrier):
				pAmericanPlayer.initUnit(self.iCarrierID, iX, iY, UnitAITypes.UNITAI_CARRIER_SEA)
			for iUnitLoop in range(iNumTransport):
				pAmericanPlayer.initUnit(self.iTransportID, iX, iY, UnitAITypes.UNITAI_ASSAULT_SEA)
				
			for iUnitLoop in range(iNumF16Fighter):
				pAmericanPlayer.initUnit(self.iF16FighterID, iX, iY, UnitAITypes.UNITAI_CARRIER_AIR)
			for iUnitLoop in range(iNumF35Fighter):
				pAmericanPlayer.initUnit(self.iF35FighterID, iX, iY, UnitAITypes.UNITAI_ATTACK_AIR)
			for iUnitLoop in range(iNumStrikeAircraft):
				pAmericanPlayer.initUnit(self.iStrikeAircraftID, iX, iY, UnitAITypes.UNITAI_ATTACK_AIR)
		
		# Set event text
		if (self.bShownAmericanReinforceText == false):
			self.bShownAmericanReinforceText = true
			self.szEventText = localText.getText("TXT_KEY_BFEA_USA_REINFORCE_TEXT", ())
			self.szResultText = localText.getText("TXT_KEY_BFEA_USA_REINFORCE_RESULT", ())
			
			self.displayEventText()
		
	def addUnitedNationsReinforcements(self, iTurn):
		
		# The amount of units spawned is determined by the Chinese score:
		fChineseFactor = gc.getGame().getPlayerScore(self.iChinaID) / (self.iUNReinforceChinaFactor * 1.0)
		
		pUnitedNationsPlayer = gc.getPlayer(self.iUnitedNationsID)
		
		# Increase coffers (can't be going bankrupt now)
		pUnitedNationsPlayer.changeGold(self.iUnitedNationsGoldAmount)
		
		# Determine force list based upon what turn it is
		
		if (iTurn < self.iUnitedNationsReinforcementThreshold + self.iTurnOffset):
			iNumInfantry =		3
			iInfantryChance =	50
			iNumPossInfantry =	2
			
			iNumInfantryIFV =	0
			iInfantryIFVChance =	25
			iNumPossInfantryIFV =	2
			
			iNumAdvancedIFV =	0
			iInfantryIFVChance =	0
			iNumPossAdvancedIFV =	0
			
			iNumModernArmor =	0
			iModernArmorChance =	50
			iNumPossModernArmor =	1
			
			iNumAdvancedArmor =	0
			iAdvancedArmorChance =	0
			iNumPossAdvancedArmor =	0
			
			iNumArtillery =		1
			iArtilleryChance =	50
			iNumPossArtillery =	1
			
			iNumMobileArtillery =		0
			iMobileArtilleryChance =	40
			iNumPossMobileArtillery =	1
			
			iNumSAMInfantry =	1
			iSAMInfantryChance =	50
			iNumPossSAMInfantry =	1
			
			iNumMobileSAM =		0
			iMobileSAMChance =	50
			iNumPossMobileSAM =	1
			
			iNumGunship =		0
			iGunshipChance =	50
			iNumPossGunship =	2
			
			iNumAdvancedGunship =		0
			iAdvancedGunshipChance =	0
			iNumPossAdvancedGunship =	0
			
			iNumJetFighter =	0
			iJetFighterChance =	50
			iNumPossJetFighter =	2
			
			iNumAdvancedFighter =		0
			iAdvancedFighterChance =	0
			iNumPossAdvancedFighter =	0
			
			iNumStrikeAircraft =	0
			iStrikeAircraftChance =	50
			iNumPossStrikeAircraft =2
			
		else:
			iNumInfantry =		2
			iInfantryChance =	50
			iNumPossInfantry =	1
			
			iNumInfantryIFV =	1
			iInfantryIFVChance =	35
			iNumPossInfantryIFV =	2
			
			iNumAdvancedIFV =	0
			iInfantryIFVChance =	50
			iNumPossAdvancedIFV =	2
			
			iNumModernArmor =	1
			iModernArmorChance =	50
			iNumPossModernArmor =	2
			
			iNumAdvancedArmor =	0
			iAdvancedArmorChance =	50
			iNumPossAdvancedArmor =	1
			
			iNumArtillery =		1
			iArtilleryChance =	50
			iNumPossArtillery =	1
			
			iNumMobileArtillery =		1
			iMobileArtilleryChance =	40
			iNumPossMobileArtillery =	1
			
			iNumSAMInfantry =		0
			iSAMInfantryChance =		50
			iNumPossSAMInfantry =		2
			
			iNumGunship =			1
			iGunshipChance =		50
			iNumPossGunship =		1
			
			iNumAdvancedGunship =		0
			iAdvancedGunshipChance =	50
			iNumPossAdvancedGunship =	1
			
			iNumJetFighter =		1
			iJetFighterChance =		30
			iNumPossJetFighter =		1
			
			iNumAdvancedFighter =		0
			iAdvancedFighterChance =	30
			iNumPossAdvancedFighter =	1
			
			iNumStrikeAircraft =		0
			iStrikeAircraftChance =		50
			iNumPossStrikeAircraft =	2
			
		# Determine the number of units to spawn
		for iUnitNumLoop in range(iNumPossInfantry):
			if (self.getRand(99) < iInfantryChance):
				iNumInfantry += 1
		for iUnitNumLoop in range(iNumPossInfantryIFV):
			if (self.getRand(99) < iInfantryIFVChance):
				iNumInfantryIFV += 1
		for iUnitNumLoop in range(iNumPossAdvancedIFV):
			if (self.getRand(99) < iAdvancedIFVChance):
				iNumAdvancedIFV += 1
		for iUnitNumLoop in range(iNumPossModernArmor):
			if (self.getRand(99) < iModernArmorChance):
				iNumModernArmor += 1
		for iUnitNumLoop in range(iNumPossAdvancedArmor):
			if (self.getRand(99) < iAdvancedArmorChance):
				iNumAdvancedArmor += 1
		for iUnitNumLoop in range(iNumPossArtillery):
			if (self.getRand(99) < iArtilleryChance):
				iNumArtillery += 1
		for iUnitNumLoop in range(iNumPossMobileArtillery):
			if (self.getRand(99) < iMobileArtilleryChance):
				iNumMobileArtillery += 1
		for iUnitNumLoop in range(iNumPossSAMInfantry):
			if (self.getRand(99) < iSAMInfantryChance):
				iNumSAMInfantry += 1
		for iUnitNumLoop in range(iNumPossMobileSAM):
			if (self.getRand(99) < iMobileSAMChance):
				iNumMobileSAM += 1
		for iUnitNumLoop in range(iNumPossGunship):
			if (self.getRand(99) < iGunshipChance):
				iNumGunship += 1
		for iUnitNumLoop in range(iNumPossAdvancedGunship):
			if (self.getRand(99) < iAdvancedGunshipChance):
				iNumAdvancedGunship += 1
		for iUnitNumLoop in range(iNumPossJetFighter):
			if (self.getRand(99) < iJetFighterChance):
				iNumJetFighter += 1
		for iUnitNumLoop in range(iNumPossAdvancedFighter):
			if (self.getRand(99) < iAdvancedFighterChance):
				iNumAdvancedFighter += 1
		for iUnitNumLoop in range(iNumPossStrikeAircraft):
			if (self.getRand(99) < iStrikeAircraftChance):
				iNumStrikeAircraft += 1
		
		# Determine a plot for land units to go to, then add them
		
		iLandPlotRange = 3
		
		aiLandUnitsPlot = self.findUnitPlacementPlot(self.iUnitedNationsID, self.aiUnitedNationsReinforceLandPlot[0], self.aiUnitedNationsReinforceLandPlot[1])
		
		if (aiLandUnitsPlot == "Oh Snap"):
			return
			
		else:
			iX = aiLandUnitsPlot[0]
			iY = aiLandUnitsPlot[1]
			
			for iUnitLoop in range(iNumInfantry):
				pUnitedNationsPlayer.initUnit(self.iInfantryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumInfantryIFV):
				pUnitedNationsPlayer.initUnit(self.iInfantryIFVID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedIFV):
				pUnitedNationsPlayer.initUnit(self.iAdvancedIFVID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumModernArmor):
				pUnitedNationsPlayer.initUnit(self.iModernArmorID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedArmor):
				pUnitedNationsPlayer.initUnit(self.iAdvancedArmorID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumArtillery):
				pUnitedNationsPlayer.initUnit(self.iArtilleryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumMobileArtillery):
				pUnitedNationsPlayer.initUnit(self.iMobileArtilleryID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumSAMInfantry):
				pUnitedNationsPlayer.initUnit(self.iSAMInfantryID, iX, iY, UnitAITypes.UNITAI_COUNTER)
			for iUnitLoop in range(iNumMobileSAM):
				pUnitedNationsPlayer.initUnit(self.iMobileSAMID, iX, iY, UnitAITypes.UNITAI_COUNTER)
			for iUnitLoop in range(iNumGunship):
				pUnitedNationsPlayer.initUnit(self.iGunshipID, iX, iY, UnitAITypes.UNITAI_ATTACK)
			for iUnitLoop in range(iNumAdvancedGunship):
				pUnitedNationsPlayer.initUnit(self.iAdvancedGunshipID, iX, iY, UnitAITypes.UNITAI_ATTACK)
				
		# Add aircraft
		aiValidCities = [	self.iIMPHAL,
					self.iDHAKA	]
		
		iCityID = self.determineValidCities(self.iUnitedNationsID, aiValidCities)
		
		if (iCityID > -1):

			iCityX = pUnitedNationsPlayer.getCity(iCityID).getX()
			iCityY = pUnitedNationsPlayer.getCity(iCityID).getY()
			
			for iUnitLoop in range(iNumJetFighter):
				pUnitedNationsPlayer.initUnit(self.iJetFighterID, iCityX, iCityY, UnitAITypes.UNITAI_ATTACK_AIR)
			for iUnitLoop in range(iNumAdvancedFighter):
				pUnitedNationsPlayer.initUnit(self.iAdvancedFighterID, iCityX, iCityY, UnitAITypes.UNITAI_ATTACK_AIR)
			for iUnitLoop in range(iNumStrikeAircraft):
				pUnitedNationsPlayer.initUnit(self.iStrikeAircraftID, iCityX, iCityY, UnitAITypes.UNITAI_ATTACK_AIR)
		
		# Set event text
		if (self.bShownUnitedNationReinforceText == false):
			self.bShownUnitedNationReinforceText = true
			self.szEventText = localText.getText("TXT_KEY_BFEA_UNITED_NATIONS_REINFORCE_TEXT", ())
			self.szResultText = localText.getText("TXT_KEY_BFEA_UNITED_NATIONS_REINFORCE_RESULT", ())
			
			self.displayEventText()
		
###########################################################################################
#################################### UTILITY FUNCTIONS ####################################
###########################################################################################
		
	def getRand(self, iNum):
		
		return CyGame().getSorenRandNum(iNum, "BFEAScenario")
		
	def getScore(self, iPlayerID):
		
		iPopScore = 0
		iVPScore = 0
		iBonusScore = 0
		iUnitScore = 0
		
		# Loop through all players' cities
		pPlayer = PyPlayer(iPlayerID)
		apCityList = pPlayer.getCityList()
		
		for pCity in apCityList:
			
			# Add city all city sizes to score
			iPopScore += pCity.getPopulation()
			
			iCityID = self.getCityIDFromPyPointer(pCity)
			
			# If this is a special city, add more to the score
			if (iCityID > -1):
				
				iVPScore += 20
		
		# Per-player modifiers
		if (iPlayerID == self.iUSAID):
			iBonusScore += 60
		elif (iPlayerID == self.iJapanID):
			iBonusScore += 150
			
#		print("Player %d Score: %d, %d, %d, %d" %(iPlayerID, iPopScore, iVPScore, iBonusScore, iUnitScore))
		return (iPopScore + iVPScore + iBonusScore + iUnitScore)
		
	def determineValidCities(self, iPlayerID, aiValidCities):
		
		pPlayer = PyPlayer(iPlayerID)
		
		# Loop through Player's cities and see if there are any valid ones
		
		for iValidCity in aiValidCities:
			
			apCityList = pPlayer.getCityList()
			for pCity in apCityList:
				iCityID = pCity.getID()
				
				if (self.getCityIDFromPyPointer(pCity) == iValidCity):
					
					return iCityID
					
		return -1
		
	def findUnitPlacementPlot(self, iPlayerID, iPlotX, iPlotY, bLand = true, bIncludePlot = true, iRange = 3):
		
		aTHEPlot = []
		aiPossiblePlots = []
		iNumPossiblePlots = 0
		aiTempPlotHolder = []
		
		pThisPlayer = gc.getPlayer(iPlayerID)
		pThisTeam = gc.getTeam(pThisPlayer.getTeam())
		
		# Loop through plots in usable range
		for iXLoop in range(iPlotX - iRange, iPlotX + iRange):
			for iYLoop in range(iPlotY - iRange, iPlotY + iRange):
				
				# Map bounds
				if (iXLoop >= 0 and iXLoop < CyMap().getGridWidth() and iYLoop >= 0 and iYLoop < CyMap().getGridHeight()):
					
					# Check to see if center plot should be excluded
					if (bIncludePlot or (not bIncludePlot and (iXLoop != iPlotX) and (iYLoop != iPlotY))):
						
						# Don't include enemy city plots!
						pPlot = CyMap().plot(iXLoop, iYLoop)
						if (pPlot.isEnemyCity(pThisPlayer.getTeam()) == false):
							
							# Check whether plot type is available
							if ((bLand == true and CyMap().plot(iXLoop, iYLoop).isWater() == false) or \
							    (bLand == false and CyMap().plot(iXLoop, iYLoop).isWater() == true)):
								aiTempPlotHolder = [iXLoop, iYLoop]
								aiPossiblePlots.append(aiTempPlotHolder)
								iNumPossiblePlots += 1
					
		if (iNumPossiblePlots == 0):
			return "Oh Snap"
		
		# Now that our list of possible plots is made, we pick one of them randomly and shove aside any enemy units already there
		iChosenPlot = self.getRand(iNumPossiblePlots - 1)
		aTHEPlot = aiPossiblePlots[iChosenPlot]
		
		# Check to see if there are any other units already on this plot
		pPlot = CyMap().plot(aTHEPlot[0], aTHEPlot[1])
		
		if (pPlot.getNumUnits() > 0):
			
			bEnemiesPresent = true
			
			while(pPlot.getNumUnits() > 0 and bEnemiesPresent):
				pLoopUnit = pPlot.getUnit(0)
				pUnitOwner = gc.getPlayer(pLoopUnit.getOwner())
				
				if (pUnitOwner):
					# Check to see if this unit belongs to a player that 'our' player is at war with
					if (pThisTeam.isAtWar(pUnitOwner.getTeam())):
						# Move this enemy unit
						self.moveUnitAside(pLoopUnit)
					else:
						# If it's a friendly, don't move it!
						bEnemiesPresent = false
		
		return aTHEPlot
		
	def moveUnitAside(self, pUnitToMove):
		
		iUnitX = pUnitToMove.getX()
		iUnitY = pUnitToMove.getY()
		
		# Don't look further than 10 plots away
		iMaxRange = 10
		iRing = 1
		
		aaiTestedList = [iUnitX, iUnitY]
		
		pPlayer = gc.getPlayer(pUnitToMove.getOwner())
		pTeam = gc.getTeam(pPlayer.getTeam())
		
		# Loop in rings around the unit's plot until we find another plot to move this unit to
		while (iRing < iMaxRange):
			
			aiTempList = []
			aaiToTestList = []
			iNumToTest = 0
			aaiValidPlotList = []
			iNumValid = 0
			
			# Find plots in this ring range
			for iX in range(-iRing, iRing + 1):
				for iY in range(-iRing, iRing + 1):
					aiTempList = [iX + iUnitX, iY + iUnitY]
					
					# Don't test this plot more than once
					if (aiTempList not in aaiTestedList):
						
						pPlot = CyMap().plot(aiTempList[0], aiTempList[1])
						bValidPlot = false
						
						if (pPlot.isWater() and pUnitToMove.getDomainType() == DomainTypes.DOMAIN_SEA):
							bValidPlot = true
						elif (pPlot.isWater() == false):
							if (pUnitToMove.getDomainType() == DomainTypes.DOMAIN_LAND or pUnitToMove.getDomainType() == DomainTypes.DOMAIN_AIR):
								bValidPlot = true
						
						# Make sure this unit can move to this plot type (no land units on water, etc.)
						if (bValidPlot):
							aaiToTestList.append(aiTempList)
							iNumToTest += 1
					
			# Now test these plots
			for iTestLoop in range(iNumToTest):
				
				pPlot = CyMap().plot(iUnitX, iUnitY)
				
				iX = aaiToTestList[iTestLoop][0]
				iY = aaiToTestList[iTestLoop][1]
				
				# If there are no enemy units here then add to the final valid plot list
				if (not self.isEnemyUnits(pTeam, iX, iY)):
					
					aaiValidPlotList.append(aaiToTestList[iTestLoop])
					iNumValid += 1
					
				# Add this plot to the tested list (to prevent it being looked at again) and look at the next in this ring
				aaiTestedList.append(aaiToTestList[iTestLoop])
				
			# This ring has been iterated through and placed in a final list - now pick a random plot from that list and exit the function
			
			if (iNumValid > 0):
#				print(aaiValidPlotList)
				iChosenPlot = self.getRand(iNumValid - 1)
				
				iX = aaiValidPlotList[iChosenPlot][0]
				iY = aaiValidPlotList[iChosenPlot][1]
				
				pUnitToMove.setXY(iX, iY)
				
				return
				
			# None of the plots in this ring are valid, so time to try the next ring
			
			iRing += 1
		
	def isEnemyUnits(self, pThisTeam, iPlotX, iPlotY):
		
		# Check to see if there are any other units already on this plot
		pPlot = CyMap().plot(iPlotX, iPlotY)
		
		if (pPlot.getNumUnits() > 0):
			
			for iUnitLoop in range(pPlot.getNumUnits()):
				
				pLoopUnit = pPlot.getUnit(iUnitLoop)
				
				pUnitOwner = gc.getPlayer(pLoopUnit.getOwner())
				
				# Check to see if this unit belongs to a player that 'our' player is at war with
				if (pThisTeam.isAtWar(pUnitOwner.getTeam())):
					return true
		
		return false
		
###### CITY SCRIPT DATA ######
		
	def getCityID(self, iPlayer, iCity):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - CityID
		szScriptData = pickle.loads(pCity.getScriptData())
		iCityID = szScriptData[0]
		
		return iCityID
		
	def setCityID(self, iPlayer, iCity, iValue):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - City ID
		szScriptData = pickle.loads(pCity.getScriptData())
		iSpawnUnitCounter = szScriptData[1]
		
		# Save Script Data - City ID
		szScriptData = [iValue, iSpawnUnitCounter]
		pCity.setScriptData(pickle.dumps(szScriptData))
	
	def getCityIDFromPyPointer(self, pCity):
		
		# Load Script Data - CityID
		szScriptData = pickle.loads(pCity.getScriptData())
		iCityID = szScriptData[0]
		
		return iCityID
		
	def getCitySpawnUnitCounter(self, iPlayer, iCity):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		szScriptData = pickle.loads(pCity.getScriptData())
		iSpawnUnitCounter = szScriptData[1]
		
		return iSpawnUnitCounter
		
	def setCitySpawnUnitCounter(self, iPlayer, iCity, iValue):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		szScriptData = pickle.loads(pCity.getScriptData())
		iCityID = szScriptData[0]
		
		# Save Script Data - SpawnUnitCounter
		szScriptData = [iCityID, iValue]
		pCity.setScriptData(pickle.dumps(szScriptData))
		
	def changeCitySpawnUnitCounter(self, iPlayer, iCity, iChange):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		szScriptData = pickle.loads(pCity.getScriptData())
		iCityID = szScriptData[0]
		iSpawnUnitCounter = szScriptData[1]
		
		iSpawnUnitCounter += iChange
		
		# Save Script Data - SpawnUnitCounter
		szScriptData = [iCityID, iSpawnUnitCounter]
		pCity.setScriptData(pickle.dumps(szScriptData))
		
	### EVENTS ###
		
	def displayEventText(self):
		
		if (gc.getPlayer(CyGame().getActivePlayer()).isAlive()):
			
			szTitle = self.szGameDate = CyGameTextMgr().getTimeStr(CyGame().getGameTurn(), false)
			
			popup = PyPopup.PyPopup(-1)
			popup.setHeaderString(szTitle)
			popup.setBodyString(self.szEventText + "\n\n" + self.szResultText)
			popup.launch(true, PopupStates.POPUPSTATE_QUEUED)
		
	def addPopup(self, szText):
		
		szTitle = self.szGameDate = CyGameTextMgr().getTimeStr(CyGame().getGameTurn(), false)
		
		popup = PyPopup.PyPopup(-1)
		popup.setHeaderString(szTitle)
		popup.setBodyString(szText)
		popup.launch(true, PopupStates.POPUPSTATE_QUEUED)
		
###########################################################################################
##################################### EVENT OVERRIDES #####################################
###########################################################################################
	
	def onLoadGame(self, argsList):
		'Called when game is loaded'
		
		self.initValues()
	
	def onGameStart(self, argsList):
		'Called at the start of the game'
		
		self.setupGame()
		
		for iPlayer in range(gc.getMAX_PLAYERS()):
			player = gc.getPlayer(iPlayer)
			if (player.isAlive() and player.isHuman()):
				popupInfo = CyPopupInfo()
				popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON_SCREEN)
				popupInfo.setText(u"showDawnOfMan")
				popupInfo.addPopup(iPlayer)
		
	def onBeginGameTurn(self, argsList):
		'Called at the beginning of the end of each turn'
		iGameTurn = argsList[0]
		
		# Battle for East Asia events
		self.turnChecker(iGameTurn)
			
	def onCityAcquired(self, argsList):
		'City Acquired'
		
		owner,playerType,city,bConquest,bTrade = argsList
#		CvUtil.pyPrint('City Acquired Event: %s' %(city.getName()))
		
		self.initCity(city)
